"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const memory_fs_1 = __importDefault(require("memory-fs"));
const webpack_1 = __importDefault(require("webpack"));
const webpack_virtual_modules_1 = __importDefault(require("webpack-virtual-modules"));
const Logger_1 = require("../Logger");
const entry_1 = require("./templates/entry");
const designDependencies_1 = require("./templates/designDependencies");
const configuration_1 = require("./configuration");
const ProgressWatcher_1 = require("./ProgressWatcher");
const WatchHandle_1 = require("./WatchHandle");
exports.WatchHandle = WatchHandle_1.WatchHandle;
const logger = new Logger_1.Logger("framer-build-library:compiler");
var BundleChunkName;
(function (BundleChunkName) {
    BundleChunkName["IndexJs"] = "index.js";
    BundleChunkName["VendorsJs"] = "vendors.js";
})(BundleChunkName = exports.BundleChunkName || (exports.BundleChunkName = {}));
class Compiler {
    constructor({ root, enableBundleSplit = true, externals = [] }) {
        this.outputFileSystem = new memory_fs_1.default();
        this.progressWatcher = new ProgressWatcher_1.ProgressWatcher();
        const config = configuration_1.createConfig({
            root,
            plugins: [
                new webpack_virtual_modules_1.default({
                    [Compiler.entryName]: entry_1.entry,
                    [Compiler.designDependenciesName]: designDependencies_1.designDependencies,
                }),
                this.progressWatcher.webpackPlugin(),
            ],
            enableBundleSplit,
            externals,
            entry: Compiler.entryName,
        });
        this.compiler = webpack_1.default(config);
        this.compiler.outputFileSystem = this.outputFileSystem;
        this.compiler.hooks.compilation.tap("framer", compilation => {
            const template = compilation.mainTemplate;
            template.hooks.require.tap("save module call", (source) => {
                // We replace the normal webpack module call by wrapping it and capturing any errors.
                // A single error will not break all components. And components that are no longer
                // available can read the `error` exported and inform the user.
                return source.replace("modules[moduleId].call(module.exports, module, module.exports, __webpack_require__);", "try { modules[moduleId].call(module.exports, module, module.exports, __webpack_require__); } catch (error) { module.exports = { error } }");
            });
            template.hooks.requireExtensions.tap("save module call", (source) => {
                // We install a module called "framer/resource" that captures the current module.i if available.
                // eslint-disable-next-line @typescript-eslint/no-use-before-define
                return source + resourceUrlModuleCode;
            });
            // Clean up module id’s, so it won’t include a crazy long relative path.
            // This also ensures that the module id matching we use for framer/resource works
            compilation.hooks.beforeModuleIds.tap("clean module ids", modules => {
                // Current webpack types don’t have types for Module ;(
                for (const module of modules) {
                    const i = module.id ? module.id.indexOf("framer-package-loader.js!") : -1;
                    if (i > 0) {
                        module.id = module.id.substring(i);
                    }
                }
            });
        });
    }
    build() {
        // Several cases to handle multiple calls coming in
        // 1. A call comes in and there is already an in progress build
        //    and a second build call has already been made. Return the
        //    promise for the already queued build.
        if (this.inProgressBuildFulfilled && this.queuedBuild) {
            return this.queuedBuild;
        }
        // 2. There is an in progress build, but no queued build. Create a
        //    new promise that is attached to the current build being complete.
        if (this.inProgressBuildFulfilled && !this.queuedBuild) {
            this.queuedBuild = this.inProgressBuildFulfilled.then(() => this.webpackRun());
            return this.queuedBuild;
        }
        // 3. No in progress build is present, so create a new build and return.
        //    Store a safe version of the build promise to make sure subsequent builds are kicked off
        //    When the in-progress build completes, clean up references and make
        //    the queued build the in progress build
        const buildPromise = this.webpackRun();
        this.inProgressBuildFulfilled = buildPromise.catch().then(() => {
            if (this.queuedBuild) {
                this.inProgressBuildFulfilled = this.queuedBuild.catch();
                this.queuedBuild = undefined;
            }
            else {
                this.inProgressBuildFulfilled = undefined;
            }
        });
        return buildPromise;
    }
    watch() {
        // don't allow to spin up a second watch process for the same compiler instance
        // until the previous one is stopped
        if (this.watchHandle) {
            return this.watchHandle;
        }
        this.watchHandle = new WatchHandle_1.WatchHandle({ webpackCompiler: this.compiler, progressWatcher: this.progressWatcher });
        this.watchHandle.once(WatchHandle_1.WatchHandleEvent.Stop, () => {
            this.watchHandle = undefined;
        });
        return this.watchHandle;
    }
    webpackRun() {
        return new Promise((resolve, reject) => {
            this.compiler.run((err, stats) => {
                if (err) {
                    reject(err);
                    return;
                }
                if (stats.hasErrors()) {
                    logBuildErrors(stats);
                }
                if (stats.startTime && stats.endTime) {
                    logger.info(`build took ${determineBuildTime(stats)}ms`);
                }
                const buildResult = getBuildResult(this.outputFileSystem, stats);
                resolve(buildResult);
            });
        });
    }
}
exports.Compiler = Compiler;
// will be created by Project.prepareProjectDir()
Compiler.templateModulesFolderName = "__framer-virtual-module-templates__";
Compiler.entryName = `./entry.js`; // Needs to start with ./
Compiler.designDependenciesName = `./designDependencies.js`; // Needs to start with ./
function getBuildResult(fileSystem, stats) {
    const statsJson = stats.toJson({
        all: false,
        assets: true,
        cachedAssets: true,
    });
    const chunks = {};
    const chunkNames = Object.values(BundleChunkName);
    for (const chunkName of chunkNames) {
        const chunk = statsJson.assets.find(asset => asset.name === chunkName);
        if (!chunk)
            continue;
        chunks[chunkName] = {
            name: chunkName,
            content: () => readChunkContent(fileSystem, chunkName),
            size: chunk ? chunk.size : 0,
            isModified: chunk && chunk.emitted !== undefined ? chunk.emitted : true,
        };
    }
    return {
        chunks,
        dependencies: analyzeDependencies(stats),
        time: determineBuildTime(stats),
        hash: stats.hash,
    };
}
exports.getBuildResult = getBuildResult;
function readChunkContent(outputFileSystem, chunkName) {
    const filePath = path_1.default.join("/", chunkName);
    try {
        return outputFileSystem.readFileSync(filePath, "utf8");
    }
    catch (error) {
        if (chunkName === "index.js" || error.code !== "ENOENT") {
            logger.error(`Cannot read script ${chunkName}`, error);
        }
        return undefined;
    }
}
function logBuildErrors(stats) {
    const statsJson = stats.toJson({
        all: false,
        errors: true,
    });
    logger.error("There were build errors:");
    statsJson.errors.forEach(errorMessage => {
        logger.error(errorMessage);
    });
    statsJson.warnings.forEach(warningMessage => {
        logger.warn(warningMessage);
    });
}
function determineBuildTime({ startTime, endTime }) {
    return !startTime || !endTime ? null : endTime - startTime;
}
exports.determineBuildTime = determineBuildTime;
function analyzeDependencies(stats) {
    // Find all the dependencies that have been bundled
    // 1. The module is not included in the known list of externals
    // 2. The module has a rawRequest value (some modules have an any value)
    // 3. The rawRequest value starts with either an @ symbol or letter, per npm package name rules
    const bundledDependencies = stats.compilation.modules.filter(module => !module.external && module.rawRequest && /^[@a-z]/i.test(module.rawRequest));
    // Store the raw names of the bundled dependencies
    const bundledDependencyNames = bundledDependencies.map(({ rawRequest }) => rawRequest);
    // Filter out bundled sub-dependencies by checking if every reason the module
    // is bundled is because of another dependency to know which dependencies
    // should be installed in the project
    const missingDependencyNames = bundledDependencies
        .filter(module => !module.reasons.every((reason) => /node_modules/.test(reason.module.context)))
        .map(({ rawRequest }) => rawRequest);
    return {
        bundled: bundledDependencyNames,
        missing: missingDependencyNames,
    };
}
/* eslint-disable */
// This is the code injected into the webpack module loading logic, so packages can find out their
// installed path.
// Notice we mask our access to "module" by eval, otherwise webpack installs some build-in module.js.
const resourceUrlModuleCode = `

// asset url
var __module_i = eval(\"typeof module !== 'undefined' ? module.i : ''\");
var __framer_package = (/(node_modules\\/.*)\\/(?:build|dist)\\/index\\.js$/.exec(__module_i) || [])[1];
function __asset_url__(src) { return __WEBPACK_EXTERNAL_MODULE_framer__.serverURL(__framer_package, src) };
installedModules['framer/resource'] = { i: 'framer/resource', l: true, exports: { url: __asset_url__ } };
`;
