"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const Logger_1 = require("../Logger");
const Compiler_1 = require("./Compiler");
const logger = new Logger_1.Logger("framer-build-library:compiler");
var WatchHandleEvent;
(function (WatchHandleEvent) {
    WatchHandleEvent["Build"] = "build";
    WatchHandleEvent["Stop"] = "stop";
    WatchHandleEvent["ProgressStart"] = "progressStart";
    WatchHandleEvent["ProgressComplete"] = "progressComplete";
})(WatchHandleEvent = exports.WatchHandleEvent || (exports.WatchHandleEvent = {}));
class WatchHandle extends events_1.EventEmitter {
    constructor({ webpackCompiler, progressWatcher, }) {
        super();
        this.onProgressStart = () => {
            this.emit(WatchHandleEvent.ProgressStart);
        };
        this.onProgressComplete = () => {
            this.emit(WatchHandleEvent.ProgressComplete);
        };
        // Don't ignore node_modules or package.json here,
        // because that is what triggers the watcher when packages are installed, removed or updated
        const watchOptions = {
            aggregateTimeout: 100,
        };
        this.progressWatcher = progressWatcher;
        this.progressWatcher.on("start", this.onProgressStart);
        this.progressWatcher.on("complete", this.onProgressComplete);
        this.webpackWatcher = webpackCompiler.watch(watchOptions, async (error, stats) => {
            if (error) {
                logger.error(error);
                this.emit(WatchHandleEvent.Build, error);
                return;
            }
            if (stats.hasErrors()) {
                // https://webpack.js.org/configuration/stats/
                logger.error(stats.toString("minimal"));
            }
            if (stats.startTime && stats.endTime) {
                logger.info(`(re-)build took ${Compiler_1.determineBuildTime(stats)}ms`);
            }
            const buildResult = Compiler_1.getBuildResult(webpackCompiler.outputFileSystem, stats);
            this.emit(WatchHandleEvent.Build, undefined, buildResult);
        });
    }
    invalidate() {
        if (!this.webpackWatcher)
            return;
        this.webpackWatcher.invalidate();
    }
    stop(onStop) {
        const handleStop = () => {
            this.webpackWatcher = undefined;
            this.progressWatcher.off("start", this.onProgressStart);
            this.progressWatcher.off("complete", this.onProgressComplete);
            this.emit(WatchHandleEvent.Stop);
            this.removeAllListeners();
            if (onStop)
                onStop();
        };
        if (!this.webpackWatcher) {
            handleStop();
            return;
        }
        this.webpackWatcher.close(handleStop);
    }
    emit(type, ...args) {
        return super.emit(type, ...args);
    }
    on(type, listener) {
        return super.on(type, listener);
    }
    once(type, listener) {
        return super.once(type, listener);
    }
}
exports.WatchHandle = WatchHandle;
