"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_1 = __importDefault(require("typescript"));
const Logger_1 = require("../Logger");
const logger = new Logger_1.Logger("framer-build-library:analyzer");
const printer = typescript_1.default.createPrinter({
    newLine: typescript_1.default.NewLineKind.LineFeed,
});
function toSourceCode(node, source) {
    return printer.printNode(typescript_1.default.EmitHint.Unspecified, node, source || node.getSourceFile());
}
exports.toSourceCode = toSourceCode;
function printAllChildren(node, depth = 0) {
    logger.log(new Array(depth + 1).join("----"), typescript_1.default.formatSyntaxKind(node.kind), node.pos, node.end);
    depth++;
    node.forEachChild(c => printAllChildren(c, depth));
}
exports.printAllChildren = printAllChildren;
function referencesIdentifier(node, name) {
    let seen = false;
    function references(n) {
        if (seen)
            return;
        if (typescript_1.default.isIdentifier(n)) {
            if (n.escapedText.toString() === name) {
                seen = true;
                return;
            }
        }
        typescript_1.default.forEachChild(n, references);
    }
    references(node);
    return seen;
}
exports.referencesIdentifier = referencesIdentifier;
function hasNodeKindRange(node, firstKind, lastKind) {
    let seen = false;
    function kinds(n) {
        if (seen)
            return;
        if (n.kind >= firstKind && n.kind <= lastKind) {
            seen = true;
            return;
        }
        typescript_1.default.forEachChild(n, kinds);
    }
    kinds(node);
    return seen;
}
// at every potential repeat point, we inject a check if we might be out of budget
function createInstrument() {
    const fn = typescript_1.default.createElementAccess(typescript_1.default.createIdentifier("window"), typescript_1.default.createLiteral("__checkBudget__"));
    return typescript_1.default.createStatement(typescript_1.default.createCall(fn, [], []));
}
function instrumentIterator(node) {
    const block = node.statement;
    if (typescript_1.default.isBlock(block)) {
        return typescript_1.default.updateBlock(block, [createInstrument()].concat(block.statements));
    }
    else {
        return typescript_1.default.createBlock([createInstrument(), block]);
    }
}
let _ctx;
function instrumentVisitor(node) {
    node = typescript_1.default.visitEachChild(node, instrumentVisitor, _ctx);
    if (typescript_1.default.isWhileStatement(node)) {
        return typescript_1.default.updateWhile(node, node.expression, instrumentIterator(node));
    }
    if (typescript_1.default.isForStatement(node)) {
        return typescript_1.default.updateFor(node, node.initializer, node.condition, node.incrementor, instrumentIterator(node));
    }
    if (typescript_1.default.isFunctionDeclaration(node) && node.body) {
        const body = node.body;
        return typescript_1.default.updateFunctionDeclaration(node, node.decorators, node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, typescript_1.default.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (typescript_1.default.isFunctionExpression(node)) {
        const body = node.body;
        return typescript_1.default.updateFunctionExpression(node, node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, typescript_1.default.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (typescript_1.default.isMethodDeclaration(node) && node.body) {
        const body = node.body;
        return typescript_1.default.updateMethod(node, node.decorators, node.modifiers, node.asteriskToken, node.name, node.questionToken, node.typeParameters, node.parameters, node.type, typescript_1.default.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (typescript_1.default.isArrowFunction(node) && typescript_1.default.isBlock(node.body)) {
        const body = node.body;
        return typescript_1.default.updateArrowFunction(node, node.modifiers, node.typeParameters, node.parameters, node.type, node.equalsGreaterThanToken, typescript_1.default.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (typescript_1.default.isConstructorDeclaration(node) && node.body) {
        const body = node.body;
        const statements = body.statements;
        const first = statements[0];
        const firstToken = first ? first.getFirstToken() : null;
        const firstIsSuperCall = firstToken ? firstToken.getText() === "super" : false;
        let instrumentedBody;
        if (firstIsSuperCall) {
            instrumentedBody = [first, createInstrument()].concat(statements.slice(1));
        }
        else {
            instrumentedBody = [createInstrument()].concat(statements);
        }
        return typescript_1.default.updateConstructor(node, node.decorators, node.modifiers, node.parameters, typescript_1.default.updateBlock(body, instrumentedBody));
    }
    return node;
}
exports.instrumentVisitor = instrumentVisitor;
var EntityType;
(function (EntityType) {
    EntityType["Component"] = "component";
    EntityType["Override"] = "override";
    EntityType["Action"] = "action";
    // Device and DeviceSkin are currently not used in code, only in package.json of Framer Library, keeping for completeness
    EntityType["Device"] = "device";
    EntityType["DeviceSkin"] = "deviceSkin";
})(EntityType = exports.EntityType || (exports.EntityType = {}));
function buildInfoExport(entities) {
    const dicts = entities.map(c => {
        const properties = [typescript_1.default.createPropertyAssignment("name", typescript_1.default.createLiteral(c.name))];
        if (c.children !== undefined) {
            properties.push(typescript_1.default.createPropertyAssignment("children", typescript_1.default.createLiteral(c.children)));
        }
        properties.push(typescript_1.default.createPropertyAssignment("type", typescript_1.default.createLiteral(c.type)));
        return typescript_1.default.createObjectLiteral(properties);
    });
    const info = typescript_1.default.createVariableDeclaration("__info__", undefined, typescript_1.default.createArrayLiteral(dicts));
    return typescript_1.default.createVariableStatement([typescript_1.default.createToken(typescript_1.default.SyntaxKind.ExportKeyword)], [info]);
}
exports.buildInfoExport = buildInfoExport;
function analyzeClasses(node, entities, exports) {
    if (!typescript_1.default.isClassDeclaration(node))
        return;
    // figure out name
    let name = "Default";
    if (node.name) {
        name = node.name.getText();
    }
    if (node.modifiers) {
        for (const x of node.modifiers) {
            if (x.kind === typescript_1.default.SyntaxKind.ExportKeyword) {
                exports.add(name);
                break;
            }
        }
    }
    // must extend Component, we blindly assume that means React.Component
    let extendsComponent = false;
    if (node.heritageClauses) {
        for (const x of node.heritageClauses) {
            if (referencesIdentifier(x, "Component")) {
                extendsComponent = true;
                break;
            }
        }
    }
    if (!extendsComponent)
        return;
    // figure out if this node references this.props.children
    // there are so many ways it can, we merely look for a `children` identifier being used
    const children = referencesIdentifier(node, "children");
    const type = EntityType.Component;
    entities.set(name, { children, type });
}
function analyzeHigherOrderComponent(node, name, entities) {
    if (!node)
        return;
    if (!typescript_1.default.isCallExpression(node))
        return;
    node.arguments.forEach(n => {
        if (typescript_1.default.isIdentifier(n)) {
            const ref = n.escapedText.toString();
            if (entities.has(ref)) {
                // a call expression that direcly references a name that refers to a known component class
                entities.set(name, entities.get(ref));
            }
        }
    });
}
function analyzeAnnotatedExport(node, name, entities) {
    if (!node)
        return;
    const type = node.type;
    if (!type)
        return;
    if (type.getText() === "React.ComponentClass") {
        entities.set(name, { type: EntityType.Component });
    }
}
function analyzeOverride(node, name, entities) {
    if (!node)
        return;
    if (!node.initializer)
        return;
    const type = node.type;
    if (type &&
        typescript_1.default.isTypeReferenceNode(type) &&
        typescript_1.default.isIdentifier(type.typeName) &&
        (type.typeName.escapedText === "Override" || type.typeName.escapedText === "OverrideFunction")) {
        entities.set(name, { type: EntityType.Override });
    }
}
function analyzeOverrideFunction(node, name, entities) {
    if (!node)
        return;
    const type = node.type;
    if (type &&
        typescript_1.default.isTypeReferenceNode(type) &&
        typescript_1.default.isIdentifier(type.typeName) &&
        (type.typeName.escapedText === "Override" || type.typeName.escapedText === "OverrideFunction")) {
        entities.set(name, { type: EntityType.Override });
    }
}
function analyzeActionFunction(node, name, entities) {
    if (!node)
        return;
    const type = node.type;
    if (type &&
        typescript_1.default.isTypeReferenceNode(type) &&
        typescript_1.default.isIdentifier(type.typeName) &&
        (type.typeName.escapedText === "Action" || type.typeName.escapedText === "ActionHandler")) {
        entities.set(name, { type: EntityType.Action });
    }
}
function analyzeStatelessComponent(node, name, entities) {
    if (!node)
        return;
    let body;
    if (typescript_1.default.isFunctionDeclaration(node)) {
        body = node;
    }
    else if (node.initializer && typescript_1.default.isFunctionLike(node.initializer)) {
        body = node.initializer;
    }
    if (!body)
        return;
    // Components must start with capital, this will set them apart from most functions
    const firstChar = name[0];
    if (firstChar !== firstChar.toUpperCase())
        return;
    let isComponent = false;
    // Trust the types
    const type = node.type;
    if (type && typescript_1.default.isTypeReferenceNode(type)) {
        const typeName = type.typeName.getText();
        isComponent = typeName.match(/^(React\.)?(SFC|StatelessComponent|FunctionComponent)$/) !== null;
    }
    // Has JSX
    if (!isComponent && hasNodeKindRange(body, typescript_1.default.SyntaxKind.JsxElement, typescript_1.default.SyntaxKind.JsxExpression)) {
        isComponent = true;
    }
    if (isComponent) {
        entities.set(name, {
            children: referencesIdentifier(body, "children"),
            type: EntityType.Component,
        });
    }
}
function analyzeExportedVariableDeclaration(node, entities, exports, exported) {
    // not all nodes seem to be connected to a source file in practice
    if (!node.getSourceFile())
        return;
    node.declarations.forEach(n => {
        if (typescript_1.default.isVariableDeclaration(n)) {
            const name = n.name.getText();
            if (exported)
                exports.add(name);
            analyzeAnnotatedExport(n, name, entities);
            analyzeHigherOrderComponent(n.initializer, name, entities);
            analyzeStatelessComponent(n, name, entities);
            analyzeOverride(n, name, entities);
        }
    });
}
function analyzeExportedIdentifiers(node, exports) {
    if (typescript_1.default.isIdentifier(node)) {
        exports.add(node.escapedText.toString());
    }
    typescript_1.default.forEachChild(node, n => analyzeExportedIdentifiers(n, exports));
}
function isExported(node) {
    let exported = false;
    if (node.modifiers) {
        for (const x of node.modifiers) {
            if (x.kind === typescript_1.default.SyntaxKind.ExportKeyword) {
                exported = true;
                break;
            }
        }
    }
    return exported;
}
function analyzeExports(node, entities, exports) {
    if (typescript_1.default.isClassDeclaration(node))
        return;
    if (typescript_1.default.isVariableStatement(node)) {
        const exported = isExported(node);
        // also find higher order components if we find `ident = call(..., ident, ...)`
        analyzeExportedVariableDeclaration(node.declarationList, entities, exports, exported);
        return;
    }
    if (typescript_1.default.isFunctionDeclaration(node) && node.name && isExported(node)) {
        const name = node.name.getText();
        exports.add(name);
        analyzeStatelessComponent(node, name, entities);
        analyzeOverrideFunction(node, name, entities);
        analyzeActionFunction(node, name, entities);
        return;
    }
    if (typescript_1.default.isExportDeclaration(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
    if (typescript_1.default.isExportAssignment(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
    if (typescript_1.default.isExportSpecifier(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
}
function analyzeFile(sourceFile) {
    const statements = [];
    const exportedNames = new Set();
    const entities = new Map();
    sourceFile.statements.forEach(n => {
        if (typescript_1.default.isVariableStatement(n)) {
            if (referencesIdentifier(n, "___info__") || referencesIdentifier(n, "__info__")) {
                return;
            }
        }
        statements.push(instrumentVisitor(n));
        analyzeClasses(n, entities, exportedNames);
        analyzeExports(n, entities, exportedNames);
    });
    if (exportedNames.size > 0) {
        const results = [];
        for (const name of entities.keys()) {
            if (!exportedNames.has(name))
                continue;
            const { children, type } = entities.get(name);
            results.push({ name, children, type });
        }
        statements.push(buildInfoExport(results));
    }
    return typescript_1.default.updateSourceFileNode(sourceFile, typescript_1.default.createNodeArray(statements));
}
function analyzeAndExportComponents(ctx) {
    return (sourceFile) => {
        _ctx = ctx;
        return analyzeFile(sourceFile);
    };
}
exports.analyzeAndExportComponents = analyzeAndExportComponents;
