"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const FramerSourceMapPlugin_1 = require("./plugins/FramerSourceMapPlugin");
const analyzer_1 = require("./analyzer");
const Compiler_1 = require("./Compiler");
function createConfig({ root, entry, enableBundleSplit, plugins = [], externals = [], }) {
    if (externals.length > 0) {
        externals.unshift(/^framer-package-loader!/);
    }
    const entryModulePath = path_1.default.join(root, Compiler_1.Compiler.entryName);
    const designDependenciesModulePath = path_1.default.join(root, Compiler_1.Compiler.designDependenciesName);
    return {
        context: root,
        entry,
        devtool: false,
        cache: true,
        mode: "development",
        output: {
            path: "/",
            filename: "index.js",
            libraryTarget: "umd",
        },
        resolve: {
            extensions: [".ts", ".tsx", ".js", ".jsx", ".css"],
        },
        resolveLoader: {
            modules: [
                // allows projects to add/override loaders
                "node_modules",
                // make sure framer-package-loader can be resolved
                `${__dirname}/loaders`,
            ],
        },
        externals: [
            {
                "framer/resource": {
                    root: "framer/resource",
                    commonjs2: "framer/resource",
                    commonjs: "framer/resource",
                    amd: "framer/resource",
                },
                framer: {
                    root: "Framer",
                    commonjs2: "framer",
                    commonjs: "framer",
                    amd: "framer",
                },
                react: {
                    root: "React",
                    commonjs2: "react",
                    commonjs: "react",
                    amd: "react",
                },
                "react-dom": {
                    root: "ReactDOM",
                    commonjs2: "react-dom",
                    commonjs: "react-dom",
                    amd: "react-dom",
                },
            },
            ...externals,
        ],
        plugins: [new FramerSourceMapPlugin_1.FramerSourceMapPlugin(), ...plugins],
        optimization: enableBundleSplit
            ? {
                splitChunks: {
                    cacheGroups: {
                        vendors: {
                            test: /\/designDependencies\.js$|\/node_modules\//,
                            filename: "vendors.js",
                            chunks: "all",
                            enforce: true,
                        },
                    },
                },
            }
            : undefined,
        module: {
            // parsing this module is usually the most expensive operation of the re-build
            // it is also unnecessary because it doesn't contain any `import` statements
            noParse: modulePath => modulePath.endsWith(designDependenciesModulePath),
            rules: [
                {
                    test: modulePath => modulePath.endsWith(entryModulePath),
                    use: {
                        loader: path_1.default.join(__dirname, "./loaders/entry-loader.js"),
                    },
                },
                {
                    test: modulePath => modulePath.endsWith(designDependenciesModulePath),
                    use: {
                        loader: path_1.default.join(__dirname, "./loaders/design-dependencies-loader.js"),
                    },
                },
                {
                    test: /\.(ts|tsx)$/,
                    use: [
                        {
                            loader: require.resolve("ts-loader"),
                            options: {
                                transpileOnly: true,
                                colors: false,
                                compilerOptions: {
                                    inlineSourceMap: true,
                                },
                                getCustomTransformers: () => ({
                                    before: [analyzer_1.analyzeAndExportComponents],
                                }),
                            },
                        },
                    ],
                },
                {
                    test: /\.css$/,
                    use: [{ loader: require.resolve("style-loader") }, { loader: require.resolve("css-loader") }],
                },
            ],
        },
    };
}
exports.createConfig = createConfig;
