"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const ini_1 = __importDefault(require("ini"));
const fs_1 = require("../fs");
const Yarn_1 = require("../Yarn");
exports.DependencyTypes = Yarn_1.YarnDependencyTypes;
const Logger_1 = require("../Logger");
const FramerLibraryManager_1 = require("../FramerLibraryManager");
const logger = new Logger_1.Logger("framer-build-library:package");
class Package {
    constructor({ paths, packageJson, builtInFramerVersion, }) {
        this.paths = paths;
        this.packageJson = packageJson;
        const yarnDefaults = [
            "--cwd",
            paths.project,
            "--skip-integrity-check",
            "--cache-folder",
            paths.yarnCache,
            "--mutex",
            `file:${paths.yarnMutex}`,
            "--ignore-scripts",
            "--json",
        ];
        if (paths.yarnLinkDir) {
            yarnDefaults.push("--link-folder", paths.yarnLinkDir);
        }
        this.yarn = new Yarn_1.Yarn({ defaults: yarnDefaults });
        this.framerLibraryManager = new FramerLibraryManager_1.FramerLibraryManager({
            paths: this.paths,
            builtInFramerVersion: builtInFramerVersion,
            pkg: this,
        });
    }
    static async load({ paths, packageJson = {}, npmrcTemplate, builtInFramerVersion, }) {
        await Package.prepare({ paths, npmrcTemplate });
        const pkg = await preparePackageJson(paths, packageJson);
        return new Package({ paths, packageJson: pkg, builtInFramerVersion });
    }
    static async prepare({ paths, npmrcTemplate }) {
        await Promise.all([
            writeYarnrc(paths.yarnrc, paths.yarnLinkDir),
            writeNpmrc(paths.npmrc, npmrcTemplate),
            fs_1.removeSymbolicLink(paths.nodeModules).catch(error => logger.info(error)),
        ]);
    }
    async readPackageName(reload = false) {
        const content = await this.read(reload);
        return content.name;
    }
    async writePackageName(name) {
        const content = await this.read();
        return this.write({ ...content, name });
    }
    async readPackageVersion(reload = false) {
        const content = await this.read(reload);
        return content.version;
    }
    async writePackageVersion(version) {
        const content = await this.read();
        return this.write({ ...content, version });
    }
    async readAuthor(reload = false) {
        const content = await this.read(reload);
        return content.author;
    }
    async writeAuthor(author) {
        const content = await this.read();
        return this.write({ ...content, author });
    }
    async readDisplayName(reload = false) {
        const content = await this.read(reload);
        return content.framer.displayName;
    }
    async writeDisplayName(displayName) {
        const content = await this.read();
        return this.write({
            ...content,
            framer: {
                ...content.framer,
                displayName,
            },
        });
    }
    async install(...args) {
        this.invalidatePackageJson();
        logger.info("Installing dependencies...");
        const start = process.hrtime();
        // Install dependencies, prefering already installed
        await this.framerLibraryManager.install();
        const installResult = await this.yarn.install(...args);
        if (installResult.type === "failure") {
            throw installResult.error;
        }
        await this.check();
        const end = process.hrtime(start);
        logger.info("Installing dependencies - DONE. %ds %dms", end[0], end[1] / 1000000);
        return installResult;
    }
    async installFramer(version) {
        await this.framerLibraryManager.install(version);
        await this.check();
    }
    publish(...args) {
        return this.yarn.publish(...args);
    }
    async readDependencies(reload = false) {
        const { dependencies, devDependencies, peerDependencies } = await this.read(reload);
        return { dependencies, devDependencies, peerDependencies };
    }
    addDependency(dependency, version) {
        this.invalidatePackageJson();
        return this.yarn.add(dependency, version, Yarn_1.YarnDependencyTypes.dependency);
    }
    addPeerDependency(dependency, version) {
        this.invalidatePackageJson();
        return this.yarn.add(dependency, version, Yarn_1.YarnDependencyTypes.peerDependency);
    }
    addDevDependency(dependency, version) {
        this.invalidatePackageJson();
        return this.yarn.add(dependency, version, Yarn_1.YarnDependencyTypes.devDependency);
    }
    removeDependency(dependency) {
        this.invalidatePackageJson();
        return this.yarn.remove(dependency);
    }
    linkDependency(dependency) {
        this.invalidatePackageJson();
        return this.yarn.link(dependency);
    }
    unlinkDependency(dependency) {
        this.invalidatePackageJson();
        return this.yarn.unlink(dependency);
    }
    upgradeDependency(dependency, version) {
        this.invalidatePackageJson();
        return this.yarn.upgrade(dependency, version);
    }
    info(packageName, info) {
        return this.yarn.info(packageName, info);
    }
    async check() {
        const compatibility = await this.framerLibraryManager.libraryCompatibility();
        if (!compatibility.compatible) {
            const message = `${compatibility.message}\n\n${compatibility.recoverySuggestion}`;
            if (compatibility.isError) {
                logger.warn(message);
            }
            else {
                logger.error(message);
            }
        }
        return this.yarn.check();
    }
    async getLatestPublishedVersion() {
        const name = await this.readPackageName();
        if (!name) {
            throw new Error("Package has not been previously published.");
        }
        const { error, inspect } = await this.info(name, "version");
        if (error.length) {
            throw new Error(error.map(({ data }) => data).join("\n"));
        }
        else if (!inspect.length) {
            throw new Error("Could not find a latest published version.");
        }
        return inspect[0].data;
    }
    async read(reload = false) {
        if (this.packageJson !== undefined && !reload) {
            return this.packageJson;
        }
        const content = await fs_1.readJSON(this.paths.packageJson, true);
        this.packageJson = content;
        return this.packageJson;
    }
    async write(content) {
        await fs_1.writeJSON(this.paths.packageJson, content, true);
        this.packageJson = content;
        return this.packageJson;
    }
    invalidatePackageJson() {
        this.packageJson = undefined;
    }
}
exports.Package = Package;
// We need to write this to the yarnrc because it is not picked up when put in the .npmrc
// The link-folder command line argument is also passed by the server,
// this is only so that external yarn commands also pick the right link folder
// Otherwise running `yarn add` in the project folder would remove the linked framer library
async function writeYarnrc(yarnrcPath, yarnLinkDir) {
    const yarnrcContents = `--link-folder "${yarnLinkDir}"`;
    await fs_extra_1.default.writeFile(yarnrcPath, yarnrcContents, "utf8");
    return yarnrcPath;
}
// Handling writing npmrc for use by Yarn for Framer registry access
// If there is an existing npmrc file, we need to merge the contents to
// protect the user's potential settings for their own registries
async function writeNpmrc(npmrcPath, npmrcTemplate) {
    if (!npmrcTemplate) {
        return fs_extra_1.default.ensureFile(npmrcPath);
    }
    // Parse the existing npmrc via ini and turn back into a string
    // Merge the template npmrc with the existing npmrc content
    // Any of the new values that have matching keys will be preferred over
    // existing values that have the same key
    // This will resolve keys changing and help future-proof against
    // new key/values that might come from the API's npmrc template in the
    // future
    let existingNpmrc;
    try {
        existingNpmrc = await fs_extra_1.default.readFile(npmrcPath, "utf8");
    }
    catch (e) {
        existingNpmrc = "";
    }
    const content = [existingNpmrc, npmrcTemplate].join("\n");
    const npmrc = ini_1.default.stringify(ini_1.default.parse(content));
    return fs_extra_1.default.writeFile(npmrcPath, npmrc, "utf8");
}
// Ensure package.json is properly structured with main, framer,
// devDependencies, peerDependnecies, and dependencies properties.
// If there is an existing package.json file, we need to merge the contents
// to protect the user's potential settings
async function preparePackageJson(paths, properties) {
    const packageJson = ((await fs_1.readJSON(paths.packageJson)) || {});
    return fs_1.writeJSON(paths.packageJson, {
        license: "MIT",
        ...packageJson,
        name: properties.name || packageJson.name || "",
        version: properties.version || packageJson.version || "1.0.0",
        author: properties.author || packageJson.author || "",
        main: path_1.default.relative(paths.project, paths.main),
        dependencies: properties.dependencies || packageJson.dependencies || {},
        peerDependencies: properties.peerDependencies || packageJson.peerDependencies || {},
        devDependencies: properties.devDependencies || packageJson.devDependencies || {},
        framer: {
            displayName: (properties.framer && properties.framer.displayName) ||
                (packageJson.framer && packageJson.framer.displayName) ||
                "",
        },
    }, true);
}
