"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const os_1 = __importDefault(require("os"));
const assert_1 = __importDefault(require("assert"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const v4_1 = __importDefault(require("uuid/v4"));
const Compiler_1 = require("../Compiler");
const Package_1 = require("../Package/Package");
const Logger_1 = require("../Logger");
const Paths_1 = require("./Paths");
exports.ProjectPaths = Paths_1.ProjectPaths;
const logger = new Logger_1.Logger("framer-build-library:project");
class Project {
    constructor({ paths, pkg, compiler }) {
        this.paths = paths;
        this.path = this.paths.project;
        this.pkg = pkg;
        this.compiler = compiler;
    }
    static async load(config) {
        const { cachePath, yarnLinkDir, npmrcTemplate, builtInFramerVersion = "1.0.0" } = config;
        const projectPath = await Project.prepareDir(config);
        const paths = new Paths_1.ProjectPaths({
            projectPath,
            cachePath,
            yarnLinkDir,
        });
        const pkg = await Package_1.Package.load({ npmrcTemplate, paths, builtInFramerVersion });
        const compiler = new Compiler_1.Compiler({
            ...config.compiler,
            root: paths.project,
        });
        const project = new Project({ paths, pkg, compiler });
        await project.install();
        return project;
    }
    static async prepareDir(config) {
        const { cachePath, yarnLinkDir } = config;
        const projectPath = await Project.extract(config.projectPath);
        const paths = new Paths_1.ProjectPaths({
            projectPath,
            cachePath,
            yarnLinkDir,
        });
        await Promise.all([fs_extra_1.default.ensureDir(paths.metadata), fs_extra_1.default.ensureDir(paths.code), fs_extra_1.default.ensureDir(paths.distOutputDir)]);
        return projectPath;
    }
    static extract(projectPath) {
        let newProjectPath = projectPath;
        if (projectPath.endsWith(".framerx")) {
            // This is a compressed archive.
            const archivePath = projectPath;
            newProjectPath = path_1.default.join(os_1.default.tmpdir(), v4_1.default(), "container");
            // TODO: Consider using blacklist to ignore unnecessary stuff (like .backups).
            return new Promise((resolve, reject) => {
                // eslint-disable-next-line @typescript-eslint/no-var-requires
                const extract = require("extract-zip");
                extract(archivePath, { dir: newProjectPath }, (error) => {
                    if (error) {
                        reject(error);
                        return;
                    }
                    resolve(newProjectPath);
                });
            });
        }
        return validateIsFramerProject(newProjectPath);
    }
    async install() {
        // Install dependencies, prefering already installed
        return this.pkg.install("--prefer-offline");
    }
    installFramer(version) {
        return this.pkg.installFramer(version);
    }
    addDependency(dependency, version) {
        return this.pkg.addDependency(dependency, version);
    }
    addPeerDependency(dependency, version) {
        return this.pkg.addPeerDependency(dependency, version);
    }
    addDevDependency(dependency, version) {
        return this.pkg.addDevDependency(dependency, version);
    }
    removeDependency(dependency) {
        return this.pkg.removeDependency(dependency);
    }
    getLatestPublishedVersion() {
        return this.pkg.getLatestPublishedVersion();
    }
    build() {
        return this.compiler.build();
    }
    async buildDist() {
        // await this.prepareForPublish(packageName, displayName)
        await this.install();
        const externals = Object.keys((await this.pkg.readDependencies(true)).dependencies);
        const compiler = new Compiler_1.Compiler({ root: this.paths.project, enableBundleSplit: false, externals });
        const build = await compiler.build();
        await this.writeBuildOutputToDisk(build);
        return build;
    }
    async publish({ version, packageName, displayName, access, }) {
        const options = ["--no-git-tag-version"];
        if (version) {
            options.push("--new-version", version);
        }
        if (access === "restricted" || access === "public") {
            options.push("--access", access);
        }
        if (packageName) {
            await this.setPackageName(packageName);
        }
        if (displayName) {
            await this.setDisplayName(displayName);
        }
        await this.buildDist();
        return this.pkg.publish(...options);
    }
    watch() {
        return this.compiler.watch();
    }
    getDisplayName() {
        return this.pkg.readDisplayName();
    }
    getAuthor() {
        return this.pkg.readAuthor();
    }
    getPackageName() {
        return this.pkg.readPackageName();
    }
    getPackageVersion() {
        return this.pkg.readPackageVersion();
    }
    async getDependencies() {
        return this.pkg.readDependencies();
    }
    setDisplayName(name) {
        return this.pkg.writeDisplayName(name);
    }
    setAuthor(author) {
        return this.pkg.writeAuthor(author);
    }
    setPackageName(name) {
        return this.pkg.writePackageName(name);
    }
    setPackageVersion(version) {
        return this.pkg.writePackageVersion(version);
    }
    // Writing the in memory output from Webpack to the FS
    // requires injecting the canvas data into the index.js file
    // and writing each chunk to disk
    async writeBuildOutputToDisk(build) {
        if (!build) {
            logger.error("Attempted to write non-existing build to disk");
            return;
        }
        await Promise.all(Object.values(Compiler_1.BundleChunkName).map(fileName => {
            if (!build.chunks[fileName]) {
                return;
            }
            let fileContent = build.chunks[fileName].content();
            if (!fileContent) {
                logger.error(`There is no ${fileName} content in the build`);
                return;
            }
            if (fileName === Compiler_1.BundleChunkName.IndexJs) {
                fileContent = injectCanvasData(fileContent);
            }
            return fs_extra_1.default.writeFile(path_1.default.join(this.paths.distOutputDir, fileName), fileContent);
        })).catch(error => {
            logger.error(error);
            throw error;
        });
    }
}
exports.Project = Project;
/** Add live canvas data to the CanvasStore of a build. */
function injectCanvasData(fileContent, canvasData = '{"children":[]}') {
    // Search for the last entry, in case some external framer packages didn't replace their CANVAS_DATA
    const needle1 = "CanvasStore.shared(";
    const needle2 = "); // CANVAS_DATA;\n";
    let at = fileContent.lastIndexOf(needle1 + needle2);
    if (at < 0) {
        return fileContent;
    }
    at += needle1.length;
    // always remove the CANVAS_DATA marker
    return fileContent.slice(0, at) + canvasData + ");\n" + fileContent.slice(at + needle2.length);
}
// Sanity checks to make sure project is a valid Framer project
async function validateIsFramerProject(projectPath) {
    assert_1.default.strict((await fs_extra_1.default.stat(projectPath)).isDirectory(), "Project is not a valid Framer project");
    return projectPath;
}
